import math
from array import array as array_
from unittest.mock import patch

import numpy as np
import pytest

from auditok import signal
from auditok.exceptions import TimeFormatError
from auditok.util import (
    AudioEnergyValidator,
    make_channel_selector,
    make_duration_formatter,
)


@pytest.fixture
def setup_data():
    return b"012345679ABC"


@pytest.mark.parametrize(
    "fmt, duration, expected",
    [
        ("%S", 5400, "5400.000"),  # only_seconds
        ("%I", 5400, "5400000"),  # only_millis
        ("%h:%m:%s.%i", 3725.365, "01:02:05.365"),  # full
        ("%h:%m:%s.%i", 1925.075, "00:32:05.075"),  # full_zero_hours
        ("%h:%m:%s.%i", 3659.075, "01:00:59.075"),  # full_zero_minutes
        ("%h:%m:%s.%i", 3720.075, "01:02:00.075"),  # full_zero_seconds
        ("%h:%m:%s.%i", 3725, "01:02:05.000"),  # full_zero_millis
        (
            "%h %h:%m:%s.%i %s",
            3725.365,
            "01 01:02:05.365 05",
        ),  # duplicate_directive
        ("%h:%m:%s", 3725, "01:02:05"),  # no_millis
        ("%h:%m", 3725, "01:02"),  # no_seconds
        ("%h", 3725, "01"),  # no_minutes
        ("%m:%s.%i", 3725, "02:05.000"),  # no_hours
    ],
    ids=[
        "only_seconds",
        "only_millis",
        "full",
        "full_zero_hours",
        "full_zero_minutes",
        "full_zero_seconds",
        "full_zero_millis",
        "duplicate_directive",
        "no_millis",
        "no_seconds",
        "no_minutes",
        "no_hours",
    ],
)
def test_make_duration_formatter(fmt, duration, expected):
    formatter = make_duration_formatter(fmt)
    result = formatter(duration)
    assert result == expected


@pytest.mark.parametrize(
    "fmt",
    [
        "%S %S",  # duplicate_only_seconds
        "%I %I",  # duplicate_only_millis
        "%x",  # unknown_directive
    ],
    ids=[
        "duplicate_only_seconds",
        "duplicate_only_millis",
        "unknown_directive",
    ],
)
def test_make_duration_formatter_error(fmt):
    with pytest.raises(TimeFormatError):
        make_duration_formatter(fmt)


@pytest.mark.parametrize(
    "sample_width, channels, selected, expected",
    [
        (
            1,
            1,
            0,
            [48, 49, 50, 51, 52, 53, 54, 55, 57, 65, 66, 67],
        ),  # int8_1channel_select_0
        (1, 2, 0, [48, 50, 52, 54, 57, 66]),  # int8_2channel_select_0
        (1, 3, 0, [48, 51, 54, 65]),  # int8_3channel_select_0
        (1, 3, 1, [49, 52, 55, 66]),  # int8_3channel_select_1
        (1, 3, 2, [50, 53, 57, 67]),  # int8_3channel_select_2
        (1, 4, 0, [48, 52, 57]),  # int8_4channel_select_0
        (
            2,
            1,
            0,
            [12592, 13106, 13620, 14134, 16697, 17218],
        ),  # int16_1channel_select_0
        (2, 2, 0, [12592, 13620, 16697]),  # int16_2channel_select_0
        (2, 2, 1, [13106, 14134, 17218]),  # int16_2channel_select_1
        (2, 3, 0, [12592, 14134]),  # int16_3channel_select_0
        (2, 3, 1, [13106, 16697]),  # int16_3channel_select_1
        (2, 3, 2, [13620, 17218]),  # int16_3channel_select_2
        (
            4,
            1,
            0,
            [858927408, 926299444, 1128415545],
        ),  # int32_1channel_select_0
        (4, 3, 0, [858927408]),  # int32_3channel_select_0
        (4, 3, 1, [926299444]),  # int32_3channel_select_1
        (4, 3, 2, [1128415545]),  # int32_3channel_select_2
    ],
    ids=[
        "int8_1channel_select_0",
        "int8_2channel_select_0",
        "int8_3channel_select_0",
        "int8_3channel_select_1",
        "int8_3channel_select_2",
        "int8_4channel_select_0",
        "int16_1channel_select_0",
        "int16_2channel_select_0",
        "int16_2channel_select_1",
        "int16_3channel_select_0",
        "int16_3channel_select_1",
        "int16_3channel_select_2",
        "int32_1channel_select_0",
        "int32_3channel_select_0",
        "int32_3channel_select_1",
        "int32_3channel_select_2",
    ],
)
def test_make_channel_selector_one_channel(
    setup_data, sample_width, channels, selected, expected
):

    selector = make_channel_selector(sample_width, channels, selected)
    result = selector(setup_data)

    dtype = signal.SAMPLE_WIDTH_TO_DTYPE[sample_width]
    expected = np.array(expected).astype(dtype)
    assert (result == expected).all()


@pytest.mark.parametrize(
    "sample_width, channels, selected, expected",
    [
        (
            1,
            1,
            "avg",
            [48, 49, 50, 51, 52, 53, 54, 55, 57, 65, 66, 67],
        ),  # int8_1channel
        (1, 2, "mix", [48.5, 50.5, 52.5, 54.5, 61, 66.5]),  # int8_2channel
        (1, 4, "average", [49.5, 53.5, 63.75]),  # int8_4channel
        (
            2,
            1,
            "mix",
            [12592, 13106, 13620, 14134, 16697, 17218],
        ),  # int16_1channel
        (2, 2, "avg", [12849, 13877, 16957.5]),  # int16_2channel
        (4, 3, "average", [971214132.33]),  # int32_3channel
    ],
    ids=[
        "int8_1channel",
        "int8_2channel",
        "int8_4channel",
        "int16_1channel",
        "int16_2channel",
        "int32_3channel",
    ],
)
def test_make_channel_selector_average(
    setup_data, sample_width, channels, selected, expected
):

    selector = make_channel_selector(sample_width, channels, selected)
    result = selector(setup_data).round(2)
    assert (result == expected).all()


@pytest.mark.parametrize(
    "sample_width, channels, selected, expected",
    [
        (
            1,
            1,
            "any",
            [[48, 49, 50, 51, 52, 53, 54, 55, 57, 65, 66, 67]],
        ),  # int8_1channel
        (
            1,
            2,
            None,
            [[48, 50, 52, 54, 57, 66], [49, 51, 53, 55, 65, 67]],
        ),  # int8_2channel
        (
            1,
            4,
            "any",
            [[48, 52, 57], [49, 53, 65], [50, 54, 66], [51, 55, 67]],
        ),  # int8_4channel
        (
            2,
            2,
            None,
            [[12592, 13620, 16697], [13106, 14134, 17218]],
        ),  # int16_2channel
        (
            4,
            3,
            "any",
            [[858927408], [926299444], [1128415545]],
        ),  # int32_3channel
    ],
    ids=[
        "int8_1channel",
        "int8_2channel",
        "int8_4channel",
        "int16_2channel",
        "int32_3channel",
    ],
)
def test_make_channel_selector_any(
    setup_data, sample_width, channels, selected, expected
):

    # Use signal functions with numpy implementation
    selector = make_channel_selector(sample_width, channels, selected)
    result = selector(setup_data)
    assert (result == expected).all()


class TestAudioEnergyValidator:
    @pytest.mark.parametrize(
        "data, channels, use_channel, expected",
        [
            ([350, 400], 1, None, True),  # mono_valid_uc_None
            ([350, 400], 1, "any", True),  # mono_valid_uc_any
            ([350, 400], 1, 0, True),  # mono_valid_uc_0
            ([350, 400], 1, "mix", True),  # mono_valid_uc_mix
            ([300, 300], 1, None, False),  # mono_invalid_uc_None
            ([300, 400, 350, 300], 2, None, True),  # stereo_valid_uc_None
            ([300, 400, 350, 300], 2, "any", True),  # stereo_valid_uc_any
            ([300, 400, 350, 300], 2, "mix", True),  # stereo_valid_uc_mix
            ([300, 400, 350, 300], 2, "avg", True),  # stereo_valid_uc_avg
            (
                [300, 400, 300, 300],
                2,
                "average",
                True,
            ),  # stereo_valid_uc_average
            (
                [634, 0, 634, 0],
                2,
                "mix",
                True,
            ),  # stereo_valid_uc_mix_with_null_channel
            ([320, 100, 320, 100], 2, 0, True),  # stereo_valid_uc_0
            ([100, 320, 100, 320], 2, 1, True),  # stereo_valid_uc_1
            ([280, 100, 280, 100], 2, None, False),  # stereo_invalid_uc_None
            ([280, 100, 280, 100], 2, "any", False),  # stereo_invalid_uc_any
            ([400, 200, 400, 200], 2, "mix", False),  # stereo_invalid_uc_mix
            ([300, 400, 300, 400], 2, 0, False),  # stereo_invalid_uc_0
            ([400, 300, 400, 300], 2, 1, False),  # stereo_invalid_uc_1
            ([0, 0, 0, 0], 2, None, False),  # zeros
        ],
        ids=[
            "mono_valid_uc_None",
            "mono_valid_uc_any",
            "mono_valid_uc_0",
            "mono_valid_uc_mix",
            "mono_invalid_uc_None",
            "stereo_valid_uc_None",
            "stereo_valid_uc_any",
            "stereo_valid_uc_mix",
            "stereo_valid_uc_avg",
            "stereo_valid_uc_average",
            "stereo_valid_uc_mix_with_null_channel",
            "stereo_valid_uc_0",
            "stereo_valid_uc_1",
            "stereo_invalid_uc_None",
            "stereo_invalid_uc_any",
            "stereo_invalid_uc_mix",
            "stereo_invalid_uc_0",
            "stereo_invalid_uc_1",
            "zeros",
        ],
    )
    def test_audio_energy_validator(
        self, data, channels, use_channel, expected
    ):

        data = array_("h", data)
        sample_width = 2
        energy_threshold = 50
        validator = AudioEnergyValidator(
            energy_threshold, sample_width, channels, use_channel
        )

        if expected:
            assert validator.is_valid(data)
        else:
            assert not validator.is_valid(data)



import os
from collections import namedtuple
from tempfile import TemporaryDirectory
from unittest.mock import patch

import pytest

from auditok.cmdline_util import (
    _AUDITOK_LOGGER,
    KeywordArguments,
    initialize_workers,
    make_kwargs,
    make_logger,
)
from auditok.exceptions import ArgumentError
from auditok.workers import (
    AudioEventsJoinerWorker,
    CommandLineWorker,
    PlayerWorker,
    PrintWorker,
    RegionSaverWorker,
    StreamSaverWorker,
)

_ArgsNamespace = namedtuple(
    "_ArgsNamespace",
    [
        "input",
        "max_read",
        "analysis_window",
        "sampling_rate",
        "sample_width",
        "channels",
        "use_channel",
        "input_format",
        "output_format",
        "large_file",
        "frame_per_buffer",
        "input_device_index",
        "save_stream",
        "save_detections_as",
        "join_detections",
        "plot",
        "save_image",
        "min_duration",
        "max_duration",
        "max_silence",
        "max_leading_silence",
        "max_trailing_silence",
        "drop_trailing_silence",
        "strict_min_duration",
        "energy_threshold",
        "echo",
        "progress_bar",
        "command",
        "quiet",
        "printf",
        "time_format",
        "timestamp_format",
    ],
)


@pytest.mark.parametrize(
    "save_stream, save_detections_as, join_detections, plot, save_image, use_channel, exp_use_channel, exp_record",  # noqa: B950
    [
        # no_record_no_join
        ("stream.ogg", None, None, False, None, "mix", "mix", False),
        # no_record_plot_join
        ("stream.ogg", None, 1.0, True, None, None, None, False),
        # no_record_save_image
        ("stream.ogg", None, None, True, "image.png", None, None, False),
        # record_plot
        (None, None, None, True, None, None, None, True),
        # record_save_image
        (None, None, None, False, "image.png", None, None, True),
        # int_use_channel
        ("stream.ogg", None, None, False, None, "1", 1, False),
        # save_detections_as
        ("stream.ogg", "{id}.wav", None, False, None, None, None, False),
    ],
    ids=[
        "no_record_no_join",
        "no_record_plot",
        "no_record_save_image",
        "record_plot",
        "record_save_image",
        "int_use_channel",
        "save_detections_as",
    ],
)
def test_make_kwargs(
    save_stream,
    save_detections_as,
    join_detections,
    plot,
    save_image,
    use_channel,
    exp_use_channel,
    exp_record,
):
    args = (
        "file",
        30,
        0.01,
        16000,
        2,
        2,
        use_channel,
        "raw",
        "ogg",
        True,
        None,
        1,
        save_stream,
        save_detections_as,
        join_detections,
        plot,
        save_image,
        0.2,
        10,
        0.3,
        0,  # max_leading_silence
        None,  # max_trailing_silence
        False,
        False,
        55,
    )
    misc = (
        False,
        False,
        None,
        True,
        None,
        "TIME_FORMAT",
        "TIMESTAMP_FORMAT",
    )
    args_ns = _ArgsNamespace(*(args + misc))

    io_kwargs = {
        "input": "file",
        "max_read": 30,
        "block_dur": 0.01,
        "sampling_rate": 16000,
        "sample_width": 2,
        "channels": 2,
        "use_channel": exp_use_channel,
        "save_stream": save_stream,
        "save_detections_as": save_detections_as,
        "join_detections": join_detections,
        "audio_format": "raw",
        "export_format": "ogg",
        "large_file": True,
        "frames_per_buffer": None,
        "input_device_index": 1,
        "record": exp_record,
    }

    split_kwargs = {
        "min_dur": 0.2,
        "max_dur": 10,
        "max_silence": 0.3,
        "max_leading_silence": 0,
        "max_trailing_silence": None,
        "drop_trailing_silence": False,
        "strict_min_dur": False,
        "energy_threshold": 55,
    }

    miscellaneous = {
        "echo": False,
        "command": None,
        "progress_bar": False,
        "quiet": True,
        "printf": None,
        "time_format": "TIME_FORMAT",
        "timestamp_format": "TIMESTAMP_FORMAT",
    }

    expected = KeywordArguments(io_kwargs, split_kwargs, miscellaneous)
    kwargs = make_kwargs(args_ns)
    assert kwargs == expected


def test_make_kwargs_error():

    args = (
        "file",
        30,
        0.01,
        16000,
        2,
        2,
        1,
        "raw",
        "ogg",
        True,
        None,
        1,
        None,  # save_stream
        None,
        1.0,  # join_detections
        None,
        None,
        0.2,
        10,
        0.3,
        0,  # max_leading_silence
        None,  # max_trailing_silence
        False,
        False,
        55,
        False,
        False,
        None,
        True,
        None,
        "TIME_FORMAT",
        "TIMESTAMP_FORMAT",
    )

    args_ns = _ArgsNamespace(*args)
    expected_err_msg = "using --join-detections/-j requires "
    expected_err_msg += "--save-stream/-O to be specified."
    with pytest.raises(ArgumentError) as arg_err:
        make_kwargs(args_ns)
    assert str(arg_err.value) == expected_err_msg


def test_make_logger_stderr_and_file(capsys):
    with TemporaryDirectory() as tmpdir:
        file = os.path.join(tmpdir, "file.log")
        logger = make_logger(stderr=True, file=file)
        assert logger.name == _AUDITOK_LOGGER
        assert len(logger.handlers) == 2
        assert logger.handlers[1].stream.name == file
        logger.info("This is a debug message")
        captured = capsys.readouterr()
        assert "This is a debug message" in captured.err


def test_make_logger_None():
    logger = make_logger(stderr=False, file=None)
    assert logger is None


def test_initialize_workers_all_plus_full_stream_saver():
    with patch("auditok.cmdline_util.player_for") as patched_player_for:
        with TemporaryDirectory() as tmpdir:
            export_filename = os.path.join(tmpdir, "output.wav")
            reader, tokenizer_worker = initialize_workers(
                input="tests/data/test_16KHZ_mono_400Hz.wav",
                save_stream=export_filename,
                export_format="wave",
                save_detections_as="{id}.wav",
                join_detections=None,
                echo=True,
                progress_bar=False,
                command="some command",
                quiet=False,
                printf="abcd",
                time_format="%S",
                timestamp_format="%h:%M:%S",
            )
            reader.stop()
            assert patched_player_for.called
            assert isinstance(reader, StreamSaverWorker)
            for obs, cls in zip(
                tokenizer_worker._observers,
                [
                    RegionSaverWorker,
                    PlayerWorker,
                    CommandLineWorker,
                    PrintWorker,
                ],
            ):
                assert isinstance(obs, cls)


def test_initialize_workers_all_plus_audio_event_joiner():
    with patch("auditok.cmdline_util.player_for") as patched_player_for:
        with TemporaryDirectory() as tmpdir:
            export_filename = os.path.join(tmpdir, "output.wav")
            reader, tokenizer_worker = initialize_workers(
                input="tests/data/test_16KHZ_mono_400Hz.wav",
                save_stream=export_filename,
                export_format="wave",
                save_detections_as="{id}.wav",
                join_detections=1,
                echo=True,
                progress_bar=False,
                command="some command",
                quiet=False,
                printf="abcd",
                time_format="%S",
                timestamp_format="%h:%M:%S",
            )
            assert patched_player_for.called
            assert not isinstance(reader, StreamSaverWorker)
            for obs, cls in zip(
                tokenizer_worker._observers,
                [
                    AudioEventsJoinerWorker,
                    RegionSaverWorker,
                    PlayerWorker,
                    CommandLineWorker,
                    PrintWorker,
                ],
            ):
                assert isinstance(obs, cls)


def test_initialize_workers_no_RegionSaverWorker():
    with patch("auditok.cmdline_util.player_for") as patched_player_for:
        with TemporaryDirectory() as tmpdir:
            export_filename = os.path.join(tmpdir, "output.wav")
            reader, tokenizer_worker = initialize_workers(
                input="tests/data/test_16KHZ_mono_400Hz.wav",
                save_stream=export_filename,
                export_format="wave",
                save_detections_as=None,
                join_detections=None,
                echo=True,
                progress_bar=False,
                command="some command",
                quiet=False,
                printf="abcd",
                time_format="%S",
                timestamp_format="%h:%M:%S",
            )
            reader.stop()
            assert patched_player_for.called
            assert isinstance(reader, StreamSaverWorker)
            for obs, cls in zip(
                tokenizer_worker._observers,
                [PlayerWorker, CommandLineWorker, PrintWorker],
            ):
                assert isinstance(obs, cls)


def test_initialize_workers_no_PlayerWorker():
    with patch("auditok.cmdline_util.player_for") as patched_player_for:
        with TemporaryDirectory() as tmpdir:
            export_filename = os.path.join(tmpdir, "output.wav")
            reader, tokenizer_worker = initialize_workers(
                input="tests/data/test_16KHZ_mono_400Hz.wav",
                save_stream=export_filename,
                export_format="wave",
                save_detections_as="{id}.wav",
                join_detections=None,
                echo=False,
                progress_bar=False,
                command="some command",
                quiet=False,
                printf="abcd",
                time_format="%S",
                timestamp_format="%h:%M:%S",
            )
            reader.stop()
            assert not patched_player_for.called
            assert isinstance(reader, StreamSaverWorker)
            for obs, cls in zip(
                tokenizer_worker._observers,
                [RegionSaverWorker, CommandLineWorker, PrintWorker],
            ):
                assert isinstance(obs, cls)


def test_initialize_workers_no_CommandLineWorker():
    with patch("auditok.cmdline_util.player_for") as patched_player_for:
        with TemporaryDirectory() as tmpdir:
            export_filename = os.path.join(tmpdir, "output.wav")
            reader, tokenizer_worker = initialize_workers(
                input="tests/data/test_16KHZ_mono_400Hz.wav",
                save_stream=export_filename,
                export_format="wave",
                save_detections_as="{id}.wav",
                join_detections=None,
                echo=True,
                progress_bar=False,
                command=None,
                quiet=False,
                printf="abcd",
                time_format="%S",
                timestamp_format="%h:%M:%S",
            )
            reader.stop()
            assert patched_player_for.called
            assert isinstance(reader, StreamSaverWorker)
            for obs, cls in zip(
                tokenizer_worker._observers,
                [RegionSaverWorker, PlayerWorker, PrintWorker],
            ):
                assert isinstance(obs, cls)


def test_initialize_workers_no_PrintWorker():
    with patch("auditok.cmdline_util.player_for") as patched_player_for:
        with TemporaryDirectory() as tmpdir:
            export_filename = os.path.join(tmpdir, "output.wav")
            reader, tokenizer_worker = initialize_workers(
                input="tests/data/test_16KHZ_mono_400Hz.wav",
                save_stream=export_filename,
                export_format="wave",
                save_detections_as="{id}.wav",
                join_detections=None,
                echo=True,
                progress_bar=False,
                command="some command",
                quiet=True,
                printf="abcd",
                time_format="%S",
                timestamp_format="%h:%M:%S",
            )
            reader.stop()
            assert patched_player_for.called
            assert isinstance(reader, StreamSaverWorker)
            for obs, cls in zip(
                tokenizer_worker._observers,
                [RegionSaverWorker, PlayerWorker, CommandLineWorker],
            ):
                assert isinstance(obs, cls)


def test_initialize_workers_no_observers():
    with patch("auditok.cmdline_util.player_for") as patched_player_for:
        reader, tokenizer_worker = initialize_workers(
            input="tests/data/test_16KHZ_mono_400Hz.wav",
            save_stream=None,
            export_format="wave",
            save_detections_as=None,
            echo=True,
            progress_bar=False,
            command=None,
            quiet=True,
            printf="abcd",
            time_format="%S",
            timestamp_format="%h:%M:%S",
        )
        assert patched_player_for.called
        assert not isinstance(reader, StreamSaverWorker)
        assert len(tokenizer_worker._observers) == 1



import base64
import io
import wave
from unittest.mock import MagicMock, patch

import numpy as np
import pytest

from auditok.audio import AudioRegion
from auditok.widget import (
    _audio_to_wav_b64,
    _downsample_peaks,
    _in_notebook,
    display_interactive,
)


def test_in_notebook_no_ipython():
    """_in_notebook returns False when IPython is not available."""
    assert _in_notebook() is False


def test_in_notebook_terminal():
    """_in_notebook returns False for TerminalInteractiveShell."""
    mock_shell = MagicMock()
    mock_shell.__class__.__name__ = "TerminalInteractiveShell"
    with patch("builtins.get_ipython", return_value=mock_shell, create=True):
        assert _in_notebook() is False


def test_in_notebook_zmq():
    """_in_notebook returns True for ZMQInteractiveShell."""
    mock_shell = MagicMock()
    mock_shell.__class__.__name__ = "ZMQInteractiveShell"
    with patch("builtins.get_ipython", return_value=mock_shell, create=True):
        assert _in_notebook() is True


def test_audio_to_wav_b64():
    """_audio_to_wav_b64 produces a valid base64-encoded WAV."""
    data = b"\x00\x01" * 80
    b64 = _audio_to_wav_b64(data, sr=16000, sw=2, ch=1)
    wav_bytes = base64.b64decode(b64)
    buf = io.BytesIO(wav_bytes)
    with wave.open(buf, "rb") as wf:
        assert wf.getframerate() == 16000
        assert wf.getsampwidth() == 2
        assert wf.getnchannels() == 1
        assert wf.readframes(80) == data


def test_downsample_peaks_passthrough():
    """When n_bins >= n_samples, peaks are per-sample."""
    samples = np.array([0.0, 0.5, -0.3, 1.0])
    peaks = _downsample_peaks(samples, n_bins=10)
    assert len(peaks) == 4
    for i, s in enumerate(samples):
        assert peaks[i] == [float(s), float(s)]


def test_downsample_peaks_reduction():
    """Downsampling reduces to the requested number of bins."""
    samples = np.random.randn(10000)
    peaks = _downsample_peaks(samples, n_bins=200)
    assert len(peaks) == 200
    for mn, mx in peaks:
        assert mn <= mx


def _capture_widget_html(region, events):
    """Helper: call display_interactive and return the generated HTML string."""
    captured = []
    mock_display = MagicMock(side_effect=lambda obj: captured.append(obj))
    mock_html_cls = MagicMock(side_effect=lambda s: s)
    mock_mod = MagicMock()
    mock_mod.display = mock_display
    mock_mod.HTML = mock_html_cls
    with patch.dict("sys.modules", {"IPython.display": mock_mod}):
        display_interactive(region, events)
    assert len(captured) == 1
    return captured[0]


@pytest.fixture
def widget_html():
    """Return the HTML produced by display_interactive for a short signal."""
    data = b"\x00\x01" * 800  # 0.05s at 16kHz mono 16-bit
    region = AudioRegion(data, 16000, 2, 1)
    events = [
        AudioRegion(data[:400], 16000, 2, 1, start=0.0),
        AudioRegion(data[400:], 16000, 2, 1, start=0.025),
    ]
    return _capture_widget_html(region, events)


def test_display_interactive_html_content(widget_html):
    """display_interactive produces HTML with expected elements."""
    assert "<canvas" in widget_html
    assert "auditok" in widget_html
    assert "Play all" in widget_html
    assert "Stop" in widget_html
    assert "2 events" in widget_html
    assert "16000 Hz" in widget_html


# --- Playback discipline: only one stream at a time ---


def test_stopPlayback_increments_generation(widget_html):
    """stopPlayback must increment playGen to cancel pending async decodes."""
    assert "playGen++;" in widget_html
    assert "function stopPlayback()" in widget_html


def test_stopPlayback_nulls_onended_before_stop(widget_html):
    """stopPlayback must null onended before calling stop() to prevent the
    stale callback from clobbering currentSource after a new source starts."""
    import re

    m = re.search(
        r"function stopPlayback\(\)\s*\{(.*?)\n  \}",
        widget_html,
        re.S,
    )
    assert m, "stopPlayback not found"
    body = m.group(1)
    # onended = null must appear before .stop()
    pos_null = body.index("onended = null")
    pos_stop = body.index(".stop()")
    assert (
        pos_null < pos_stop
    ), "onended must be nulled before stop() to prevent stale callback"


def test_playB64_checks_generation(widget_html):
    """playB64 async callback must bail out when generation is stale."""
    # playB64 should capture gen before decode and check it in callback
    assert "function playB64(" in widget_html
    assert "var gen = playGen;" in widget_html
    assert "if (gen !== playGen) return;" in widget_html


def test_playFullFrom_checks_generation(widget_html):
    """playFullFrom async callback must bail out when generation is stale."""
    assert "function playFullFrom(" in widget_html
    assert "if (gen !== playGen) return;" in widget_html


def test_all_play_paths_call_stopPlayback(widget_html):
    """Every play entry point must call stopPlayback first."""
    import re

    # playB64 body should start with stopPlayback()
    m = re.search(
        r"function playB64\([^)]*\)\s*\{(.*?)\n  \}", widget_html, re.S
    )
    assert m, "playB64 not found"
    assert "stopPlayback();" in m.group(1)

    # playFullFrom body should start with stopPlayback()
    m = re.search(
        r"function playFullFrom\([^)]*\)\s*\{(.*?)\n  \}", widget_html, re.S
    )
    assert m, "playFullFrom not found"
    assert "stopPlayback();" in m.group(1)


def test_stop_button_calls_stopPlayback(widget_html):
    """The Stop button handler must call stopPlayback."""
    assert 'btnStop.addEventListener("click"' in widget_html
    # The handler body includes stopPlayback()
    idx = widget_html.index('btnStop.addEventListener("click"')
    snippet = widget_html[idx : idx + 200]
    assert "stopPlayback();" in snippet


def test_ruler_click_calls_playFullFrom(widget_html):
    """Clicking the ruler must route through playFullFrom (single-play path)."""
    assert "isInRuler(pos)" in widget_html
    assert "playFullFrom(t);" in widget_html


def test_play_all_button_calls_playFullFrom(widget_html):
    """The Play all button routes through playFullFrom and supports pause."""
    idx = widget_html.index('btnPlay.addEventListener("click"')
    snippet = widget_html[idx : idx + 500]
    assert "playFullFrom(0);" in snippet  # play from start
    assert "playFullFrom(resumeFrom);" in snippet  # resume from pause
    assert "pausePlayback();" in snippet  # pause when playing


def test_detection_click_calls_playB64(widget_html):
    """Clicking a detection must route through playB64 (which calls stop)."""
    assert "playB64(regionB64[idx]" in widget_html


def test_pause_sets_button_text(widget_html):
    """pausePlayback and setPlaying toggle button innerHTML."""
    assert "function pausePlayback()" in widget_html
    assert "function setPlaying(" in widget_html
    # Pause label
    assert "Pause" in widget_html
    # Play label (restored on stop/end)
    assert "Play all" in widget_html


def test_stopPlayback_resets_pause_state(widget_html):
    """stopPlayback must clear pausedAt so next Play starts from beginning."""
    import re

    m = re.search(
        r"function stopPlayback\(\)\s*\{(.*?)\n  \}",
        widget_html,
        re.S,
    )
    assert m
    body = m.group(1)
    assert "pausedAt = -1" in body
    assert "setPlaying(false)" in body


def test_timestamp_display_element(widget_html):
    """Widget contains a selectable timestamp span updated during playback."""
    assert "timeSpan" in widget_html
    assert "user-select:all" in widget_html
    assert "updateTimestamp(" in widget_html
    assert "formatTimePrecise(" in widget_html


def test_energy_threshold_line():
    """When energy_threshold is provided, the widget draws a dashed line."""
    data = b"\x00\x10" * 800
    region = AudioRegion(data, 16000, 2, 1)
    events = [AudioRegion(data[:400], 16000, 2, 1, start=0.0)]
    html = _capture_widget_html(region, events)
    # No threshold by default
    assert "ethNorm" in html
    assert "[null]" in html

    # With threshold
    captured = []
    mock_display = MagicMock(side_effect=lambda obj: captured.append(obj))
    mock_html_cls = MagicMock(side_effect=lambda s: s)
    mock_mod = MagicMock()
    mock_mod.display = mock_display
    mock_mod.HTML = mock_html_cls
    with patch.dict("sys.modules", {"IPython.display": mock_mod}):
        display_interactive(region, events, energy_threshold=50)
    html_eth = captured[0]
    assert "ethNorm" in html_eth
    assert "[null]" not in html_eth  # should have a numeric value
    assert "setLineDash([6, 4])" in html_eth  # dashed line



from __future__ import annotations


class ArgumentError(Exception):
    """Raised when command line arguments have invalid values."""


class TooSmallBlockDuration(ValueError):
    """Raised when block_dur results in a block_size smaller than one sample."""

    def __init__(
        self, message: str, block_dur: float, sampling_rate: int
    ) -> None:
        self.block_dur = block_dur
        self.sampling_rate = sampling_rate
        super().__init__(message)


class TimeFormatError(Exception):
    """Raised when a duration formatting directive is unknown."""


class EndOfProcessing(Exception):
    """Raised within command line script's main function to jump to
    postprocessing code."""


class AudioIOError(Exception):
    """Raised when a compressed audio file cannot be loaded or when trying
    to read from a not yet open AudioSource"""


class AudioParameterError(AudioIOError):
    """Raised when one audio parameter is missing when loading raw data or
    saving data to a format other than raw. Also raised when an audio
    parameter has a wrong value."""


class AudioEncodingError(Exception):
    """Raised if audio data can not be encoded in the provided format"""


class AudioEncodingWarning(RuntimeWarning):
    """Raised if audio data can not be encoded in the provided format
    but saved as wav.
    """



import numpy as np

AUDITOK_PLOT_THEME = {
    "figure": {"facecolor": "#482a36", "alpha": 0.2},
    "plot": {"facecolor": "#282a36"},
    "energy_threshold": {
        "color": "#e31f8f",
        "linestyle": "--",
        "linewidth": 1,
    },
    "signal": {"color": "#40d970", "linestyle": "-", "linewidth": 1},
    "detections": {
        "facecolor": "#777777",
        "edgecolor": "#ff8c1a",
        "linewidth": 1,
        "alpha": 0.75,
    },
}


def _make_time_axis(nb_samples, sampling_rate):
    sample_duration = 1 / sampling_rate
    x = np.linspace(0, sample_duration * (nb_samples - 1), nb_samples)
    return x


def _plot_line(x, y, theme, xlabel=None, ylabel=None, **kwargs):
    import matplotlib.pyplot as plt

    color = theme.get("color", theme.get("c"))
    ls = theme.get("linestyle", theme.get("ls"))
    lw = theme.get("linewidth", theme.get("lw"))
    plt.plot(x, y, c=color, ls=ls, lw=lw, **kwargs)
    plt.xlabel(xlabel, fontsize=8)
    plt.ylabel(ylabel, fontsize=8)


def _plot_detections(subplot, detections, theme):
    fc = theme.get("facecolor", theme.get("fc"))
    ec = theme.get("edgecolor", theme.get("ec"))
    ls = theme.get("linestyle", theme.get("ls"))
    lw = theme.get("linewidth", theme.get("lw"))
    alpha = theme.get("alpha")
    for start, end in detections:
        subplot.axvspan(start, end, fc=fc, ec=ec, ls=ls, lw=lw, alpha=alpha)


def plot(
    audio_region,
    scale_signal=True,
    detections=None,
    energy_threshold=None,
    show=True,
    figsize=None,
    save_as=None,
    dpi=120,
    theme="auditok",
):
    import matplotlib.pyplot as plt

    y = np.asarray(audio_region)
    if len(y.shape) == 1:
        y = y.reshape(1, -1)
    nb_subplots, nb_samples = y.shape
    sampling_rate = audio_region.sampling_rate
    time_axis = _make_time_axis(nb_samples, sampling_rate)
    if energy_threshold is not None:
        eth_log10 = energy_threshold * np.log(10) / 10
        amplitude_threshold = np.sqrt(np.exp(eth_log10))
    else:
        amplitude_threshold = None
    if detections is None:
        detections = []
    else:
        # End of detection corresponds to the end of the last sample but
        # to stay compatible with the time axis of signal plotting we want end
        # of detection to correspond to the *start* of the that last sample.
        detections = [
            (start, end - (1 / sampling_rate)) for (start, end) in detections
        ]
    if figsize is None:
        figsize = (14, 2.5 * nb_subplots)

    if theme == "auditok":
        theme = AUDITOK_PLOT_THEME

    fig = plt.figure(figsize=figsize, dpi=dpi)
    fig_theme = theme.get("figure", theme.get("fig", {}))
    fig_fc = fig_theme.get("facecolor", fig_theme.get("ffc"))
    fig_alpha = fig_theme.get("alpha", 1)
    fig.patch.set_facecolor(fig_fc)
    fig.patch.set_alpha(fig_alpha)

    plot_theme = theme.get("plot", {})
    plot_fc = plot_theme.get("facecolor", plot_theme.get("pfc"))

    if nb_subplots > 2 and nb_subplots % 2 == 0:
        nb_rows = nb_subplots // 2
        nb_columns = 2
    else:
        nb_rows = nb_subplots
        nb_columns = 1

    for sid, samples in enumerate(y, 1):
        ax = fig.add_subplot(nb_rows, nb_columns, sid)
        ax.set_facecolor(plot_fc)
        if scale_signal:
            std = samples.std()
            if std > 0:
                mean = samples.mean()
                std = samples.std()
                samples = (samples - mean) / std
                max_ = samples.max()
                plt.ylim(-1.5 * max_, 1.5 * max_)
        if amplitude_threshold is not None:
            if scale_signal and std > 0:
                amp_th = (amplitude_threshold - mean) / std
            else:
                amp_th = amplitude_threshold
            eth_theme = theme.get("energy_threshold", theme.get("eth", {}))
            _plot_line(
                [time_axis[0], time_axis[-1]],
                [amp_th] * 2,
                eth_theme,
                label="Detection threshold",
            )
            if sid == 1:
                leg = plt.legend(
                    ["Detection threshold"],
                    facecolor=plot_fc,
                    framealpha=1.0,
                    edgecolor="#555555",
                    loc="upper right",
                    fontsize=8,
                )
                for text in leg.get_texts():
                    text.set_color("#cccccc")

        signal_theme = theme.get("signal", {})
        _plot_line(
            time_axis,
            samples,
            signal_theme,
            xlabel="Time (seconds)",
            ylabel="Signal{}".format(" (scaled)" if scale_signal else ""),
        )
        detections_theme = theme.get("detections", {})
        _plot_detections(ax, detections, detections_theme)
        plt.title("Channel {}".format(sid), fontsize=10)

        plt.xticks(fontsize=8)
        plt.yticks(fontsize=8)
    plt.tight_layout()

    if save_as is not None:
        plt.savefig(save_as, dpi=dpi)
    if show:
        plt.show()



import os
import subprocess
import wave
from abc import ABCMeta, abstractmethod
from collections import namedtuple
from datetime import datetime, timedelta
from queue import Empty, Queue
from tempfile import NamedTemporaryFile
from threading import Thread

from .audio import make_silence, split
from .exceptions import AudioEncodingError, AudioEncodingWarning
from .io import _guess_audio_format
from .util import AudioReader, make_duration_formatter

_STOP_PROCESSING = "STOP_PROCESSING"
_Detection = namedtuple("_Detection", "id start end duration")


def _run_subprocess(command):
    try:
        with subprocess.Popen(
            command,
            stdin=open(os.devnull, "rb"),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        ) as proc:
            stdout, stderr = proc.communicate()
            return proc.returncode, stdout, stderr
    except Exception as exc:
        err_msg = "Couldn't export audio using command: '{}'".format(command)
        raise AudioEncodingError(err_msg) from exc


class Worker(Thread, metaclass=ABCMeta):
    def __init__(self, timeout=0.5, logger=None):
        self._timeout = timeout
        self._logger = logger
        self._inbox = Queue()
        Thread.__init__(self)

    def run(self):
        while True:
            message = self._get_message()
            if message == _STOP_PROCESSING:
                break
            if message is not None:
                self._process_message(message)
        self._post_process()

    @abstractmethod
    def _process_message(self, message):
        """Process incoming messages"""

    def _post_process(self):
        pass

    def _log(self, message):
        self._logger.info(message)

    def _stop_requested(self):
        try:
            message = self._inbox.get_nowait()
            if message == _STOP_PROCESSING:
                return True
        except Empty:
            return False

    def stop(self):
        self.send(_STOP_PROCESSING)
        self.join()

    def send(self, message):
        self._inbox.put(message)

    def _get_message(self):
        try:
            message = self._inbox.get(timeout=self._timeout)
            return message
        except Empty:
            return None


class TokenizerWorker(Worker, AudioReader):
    def __init__(self, reader, observers=None, logger=None, **kwargs):
        self._observers = observers if observers is not None else []
        self._reader = reader
        kwargs["input"] = self
        self._audio_region_gen = split(**kwargs)
        self._detections = []
        self._log_format = "[DET]: Detection {0.id} (start: {0.start:.3f}, "
        self._log_format += "end: {0.end:.3f}, duration: {0.duration:.3f})"
        super().__init__(timeout=0.2, logger=logger)

    def _process_message(self):
        pass

    @property
    def detections(self):
        return self._detections

    @property
    def reader(self):
        return self._reader

    def _notify_observers(self, message):
        for observer in self._observers:
            observer.send(message)

    def run(self):
        self._reader.open()
        start_processing_timestamp = datetime.now()
        for _id, audio_region in enumerate(self._audio_region_gen, start=1):
            timestamp = start_processing_timestamp + timedelta(
                seconds=audio_region.start
            )
            object.__setattr__(audio_region, "timestamp", timestamp)
            detection = _Detection(
                _id,
                audio_region.start,
                audio_region.end,
                audio_region.duration,
            )
            self._detections.append(detection)
            if self._logger is not None:
                message = self._log_format.format(detection)
                self._log(message)
            self._notify_observers((_id, audio_region))
        self._notify_observers(_STOP_PROCESSING)
        self._reader.close()

    def start_all(self):
        for observer in self._observers:
            observer.start()
        self.start()

    def stop_all(self):
        self.stop()
        for observer in self._observers:
            observer.stop()
        self._reader.close()

    def read(self):
        if self._stop_requested():
            return None
        else:
            return self._reader.read()

    def __getattr__(self, name):
        return getattr(self._reader, name)


class AudioDataSaverWorker(Worker):

    def __init__(
        self,
        filename,
        export_format,
        sampling_rate,
        sample_width,
        channels,
        timeout=0.2,
    ):

        super().__init__(timeout=timeout)
        self._output_filename = filename
        self._sampling_rate = sampling_rate
        self._sample_width = sample_width
        self._channels = channels

        self._export_format = _guess_audio_format(filename, export_format)
        if self._export_format is None:
            self._export_format = "wav"
        self._init_output_stream()
        self._exported = False

    @property
    def sr(self):
        return self._sampling_rate

    @property
    def sw(self):
        return self._sample_width

    @property
    def ch(self):
        return self._channels

    def _get_non_existent_filename(self):
        filename = self._output_filename + ".wav"
        i = 0
        while os.path.exists(filename):
            i += 1
            filename = self._output_filename + "({}).wav".format(i)
        return filename

    def _init_output_stream(self):
        if self._export_format != "wav":
            self._tmp_output_filename = self._get_non_existent_filename()
        else:
            self._tmp_output_filename = self._output_filename
        self._wfp = wave.open(self._tmp_output_filename, "wb")
        self._wfp.setframerate(self.sr)
        self._wfp.setsampwidth(self.sw)
        self._wfp.setnchannels(self.ch)

    @property
    def data(self):
        with wave.open(self._tmp_output_filename, "rb") as wfp:
            return wfp.readframes(-1)

    def export_audio(self):
        try:
            self._encode_export_audio()
        except AudioEncodingError as ae_error:
            raise AudioEncodingWarning(str(ae_error)) from ae_error
        return self._output_filename

    def _encode_export_audio(self):
        if self._exported:
            return self._output_filename

        if self._export_format in ("raw", "wav"):
            if self._export_format == "raw":
                self._export_raw()
            self._exported = True
            return self._output_filename
        try:
            self._export_with_ffmpeg_or_avconv()

        except AudioEncodingError:
            try:
                self._export_with_sox()
            except AudioEncodingError as exc:
                warn_msg = "Couldn't save audio data in the desired format "
                warn_msg += "'{}'.\nEither none of 'ffmpeg', 'avconv' or 'sox' "
                warn_msg += "is installed or this format is not recognized.\n"
                warn_msg += "Audio file was saved as '{}'"
                raise AudioEncodingError(
                    warn_msg.format(
                        self._export_format, self._tmp_output_filename
                    )
                ) from exc
            else:
                self._exported = True
        else:
            self._exported = True
        return self._output_filename

    def _export_raw(self):
        with open(self._output_filename, "wb") as fp:
            fp.write(self.data)

    def _export_with_ffmpeg_or_avconv(self):
        command = [
            "-y",
            "-f",
            "wav",
            "-i",
            self._tmp_output_filename,
            "-f",
            self._export_format,
            self._output_filename,
        ]
        returncode, stdout, stderr = _run_subprocess(["ffmpeg"] + command)
        if returncode != 0:
            returncode, stdout, stderr = _run_subprocess(["avconv"] + command)
            if returncode != 0:
                raise AudioEncodingError(stderr)
        return stdout, stderr

    def _export_with_sox(self):
        command = [
            "sox",
            "-t",
            "wav",
            self._tmp_output_filename,
            self._output_filename,
        ]
        returncode, stdout, stderr = _run_subprocess(command)
        if returncode != 0:
            raise AudioEncodingError(stderr)
        return stdout, stderr

    def close_output(self):
        self._wfp.close()

    def __del__(self):
        self._post_process()

        if (
            (self._tmp_output_filename != self._output_filename)
            and self._exported
            and os.path.exists(self._tmp_output_filename)
        ):
            os.remove(self._tmp_output_filename)


class StreamSaverWorker(AudioDataSaverWorker):
    def __init__(
        self,
        audio_reader,
        filename,
        export_format=None,
        cache_size_sec=0.5,
        timeout=0.2,
    ):
        self._reader = audio_reader
        super().__init__(
            filename,
            export_format,
            self._reader.sr,
            self._reader.sw,
            self._reader.ch,
            timeout=timeout,
        )

        sample_size_bytes = self._reader.sw * self._reader.ch
        self._cache_size = cache_size_sec * self._reader.sr * sample_size_bytes

        self._exported = False
        self._cache = []
        self._total_cached = 0

    def _process_message(self, data):
        self._cache.append(data)
        self._total_cached += len(data)
        if self._total_cached >= self._cache_size:
            self._write_cached_data()

    def _post_process(self):
        while True:
            try:
                data = self._inbox.get_nowait()
                if data != _STOP_PROCESSING:
                    self._cache.append(data)
                    self._total_cached += len(data)
            except Empty:
                break
        self._write_cached_data()
        self._wfp.close()

    def _write_cached_data(self):
        if self._cache:
            data = b"".join(self._cache)
            self._wfp.writeframes(data)
            self._cache = []
            self._total_cached = 0

    def open(self):
        self._reader.open()

    def close(self):
        self._reader.close()
        self.stop()

    def rewind(self):
        # ensure compatibility with AudioReader with record=True
        pass

    @property
    def data(self):
        with wave.open(self._tmp_output_filename, "rb") as wfp:
            return wfp.readframes(-1)

    def read(self):
        data = self._reader.read()
        if data is not None:
            self.send(data)
        else:
            self.send(_STOP_PROCESSING)
        return data

    def __getattr__(self, name):
        try:
            return getattr(self._reader, name)
        except AttributeError:
            return getattr(self, name)


class AudioEventsJoinerWorker(AudioDataSaverWorker):

    def __init__(
        self,
        silence_duration,
        filename,
        export_format,
        sampling_rate,
        sample_width,
        channels,
        timeout=0.2,
    ):

        super().__init__(
            filename,
            export_format,
            sampling_rate,
            sample_width,
            channels,
            timeout,
        )

        self._silence_data = make_silence(
            silence_duration, sampling_rate, sample_width, channels
        ).data
        self._first_event = True

    def _process_message(self, message):
        _, audio_event = message
        self._write_audio_event(audio_event.data)

    def _post_process(self):
        while True:
            try:
                message = self._inbox.get_nowait()
                if message != _STOP_PROCESSING:
                    _, audio_event = message
                    self._write_audio_event(audio_event.data)
            except Empty:
                break
        self._wfp.close()

    def _write_audio_event(self, data):
        if not self._first_event:
            self._wfp.writeframes(self._silence_data)
        else:
            self._first_event = False
        self._wfp.writeframes(data)


class PlayerWorker(Worker):
    def __init__(self, player, progress_bar=False, timeout=0.2, logger=None):
        self._player = player
        self._progress_bar = progress_bar
        self._log_format = "[PLAY]: Detection {id} played"
        super().__init__(timeout=timeout, logger=logger)

    def _process_message(self, message):
        _id, audio_region = message
        if self._logger is not None:
            message = self._log_format.format(id=_id)
            self._log(message)
        audio_region.play(
            player=self._player, progress_bar=self._progress_bar, leave=False
        )


class RegionSaverWorker(Worker):
    def __init__(
        self,
        filename_format,
        audio_format=None,
        timeout=0.2,
        logger=None,
        **audio_parameters,
    ):
        self._filename_format = filename_format
        self._audio_format = audio_format
        self._audio_parameters = audio_parameters
        self._debug_format = "[SAVE]: Detection {id} saved as '{filename}'"
        super().__init__(timeout=timeout, logger=logger)

    def _process_message(self, message):
        _id, audio_region = message
        filename = self._filename_format.format(
            id=_id,
            start=audio_region.start,
            end=audio_region.end,
            duration=audio_region.duration,
        )
        filename = audio_region.save(
            filename, self._audio_format, **self._audio_parameters
        )
        if self._logger:
            message = self._debug_format.format(id=_id, filename=filename)
            self._log(message)


class CommandLineWorker(Worker):
    def __init__(self, command, timeout=0.2, logger=None):
        self._command = command
        super().__init__(timeout=timeout, logger=logger)
        self._debug_format = "[COMMAND]: Detection {id} command: '{command}'"

    def _process_message(self, message):
        _id, audio_region = message
        with NamedTemporaryFile(delete=False) as file:
            filename = audio_region.save(file.name, audio_format="wav")
            command = self._command.format(file=filename)
            os.system(command)
            if self._logger is not None:
                message = self._debug_format.format(id=_id, command=command)
                self._log(message)


class PrintWorker(Worker):
    def __init__(
        self,
        print_format="{start} {end}",
        time_format="%S",
        timestamp_format="%Y/%m/%d %H:%M:%S.%f",
        timeout=0.2,
    ):

        self._print_format = print_format
        self._format_time = make_duration_formatter(time_format)
        self._timestamp_format = timestamp_format
        self.detections = []
        super().__init__(timeout=timeout)

    def _process_message(self, message):
        _id, audio_region = message
        timestamp = audio_region.timestamp
        timestamp = timestamp.strftime(self._timestamp_format)
        text = self._print_format.format(
            id=_id,
            start=self._format_time(audio_region.start),
            end=self._format_time(audio_region.end),
            duration=self._format_time(audio_region.duration),
            timestamp=timestamp,
        )
        print(text)
